// File: @openzeppelin/contracts/utils/Context.sol


// OpenZeppelin Contracts v4.4.1 (utils/Context.sol)

pragma solidity ^0.8.0;

/**
 * @dev Provides information about the current execution context, including the
 * sender of the transaction and its data. While these are generally available
 * via msg.sender and msg.data, they should not be accessed in such a direct
 * manner, since when dealing with meta-transactions the account sending and
 * paying for execution may not be the actual sender (as far as an application
 * is concerned).
 *
 * This contract is only required for intermediate, library-like contracts.
 */
abstract contract Context {
    function _msgSender() internal view virtual returns (address) {
        return msg.sender;
    }

    function _msgData() internal view virtual returns (bytes calldata) {
        return msg.data;
    }
}

// File: @openzeppelin/contracts/access/Ownable.sol


// OpenZeppelin Contracts (last updated v4.9.0) (access/Ownable.sol)

//pragma solidity ^0.8.0;


/**
 * @dev Contract module which provides a basic access control mechanism, where
 * there is an account (an owner) that can be granted exclusive access to
 * specific functions.
 *
 * By default, the owner account will be the one that deploys the contract. This
 * can later be changed with {transferOwnership}.
 *
 * This module is used through inheritance. It will make available the modifier
 * `onlyOwner`, which can be applied to your functions to restrict their use to
 * the owner.
 */
abstract contract Ownable is Context {
    address private _owner;

    event OwnershipTransferred(address indexed previousOwner, address indexed newOwner);

    /**
     * @dev Initializes the contract setting the deployer as the initial owner.
     */
    constructor() {
        _transferOwnership(_msgSender());
    }

    /**
     * @dev Throws if called by any account other than the owner.
     */
    modifier onlyOwner() {
        _checkOwner();
        _;
    }

    /**
     * @dev Returns the address of the current owner.
     */
    function owner() public view virtual returns (address) {
        return _owner;
    }

    /**
     * @dev Throws if the sender is not the owner.
     */
    function _checkOwner() internal view virtual {
        require(owner() == _msgSender(), "Ownable: caller is not the owner");
    }

    /**
     * @dev Leaves the contract without owner. It will not be possible to call
     * `onlyOwner` functions. Can only be called by the current owner.
     *
     * NOTE: Renouncing ownership will leave the contract without an owner,
     * thereby disabling any functionality that is only available to the owner.
     */
    function renounceOwnership() public virtual onlyOwner {
        _transferOwnership(address(0));
    }

    /**
     * @dev Transfers ownership of the contract to a new account (`newOwner`).
     * Can only be called by the current owner.
     */
    function transferOwnership(address newOwner) public virtual onlyOwner {
        require(newOwner != address(0), "Ownable: new owner is the zero address");
        _transferOwnership(newOwner);
    }

    /**
     * @dev Transfers ownership of the contract to a new account (`newOwner`).
     * Internal function without access restriction.
     */
    function _transferOwnership(address newOwner) internal virtual {
        address oldOwner = _owner;
        _owner = newOwner;
        emit OwnershipTransferred(oldOwner, newOwner);
    }
}

// File: MTPLRegistry.sol

// SPDX-License-Identifier: MIT
//pragma solidity ^0.8.0;


contract Kernel is Ownable {
    struct MarketPair {
        address loanToken;
        address collateralToken;
        address marketPair;
    }

    MarketPair[] public marketPairs;

    // Mapping from pair token address to an array of indexes in marketPairs[]
    mapping(address => uint256[]) public loanTokenToMarketPairs;
    mapping(address => uint256[]) public collateralTokenToMarketPairs;

    event PairAdded(address indexed loanToken, address indexed collateralToken, address marketPair, uint _pid);

    function addPair(
        address loanToken,
        address collateralToken,
        address marketPairAddress
    ) external onlyOwner returns (uint256 _id) {
        _id = marketPairs.length;
        marketPairs.push(
            MarketPair({loanToken: loanToken, collateralToken: collateralToken, marketPair: marketPairAddress})
        );

        loanTokenToMarketPairs[loanToken].push(_id);
        collateralTokenToMarketPairs[collateralToken].push(_id);

        emit PairAdded(loanToken, collateralToken, marketPairAddress, _id);
    }

    function getByLoanToken(address loanToken) external view returns (MarketPair[] memory) {
        uint256[] memory pairIndexes = loanTokenToMarketPairs[loanToken];
        return getMarketPairsByIndexes(pairIndexes);
    }

   

    function getMarketPairsByIndexes(uint[] memory indexes) private view returns (MarketPair[] memory) {
        MarketPair[] memory pairs = new MarketPair[](indexes.length);
        for (uint i = 0; i < indexes.length; ++i) {
            pairs[i] = marketPairs[indexes[i]];
        }
        return pairs;
    }
}