// SPDX-License-Identifier: GPL-3.0

/*
    This file is part of the Enzyme Protocol.

    (c) Enzyme Council <councilenzyme.finance>

    For the full license information, please view the LICENSE
    file that was distributed with this source code.
*/

pragma solidity ^0.6.12;
pragma experimental ABIEncoderV2;

import "../utils/AddressListRegistryPolicyBase.sol";

/// title AllowedDepositRecipientsPolicy Contract
/// author Enzyme Council <securityenzyme.finance>
/// notice A policy that limits the accounts that can receive shares via deposit
contract AllowedDepositRecipientsPolicy is AddressListRegistryPolicyBase {
    constructor(address _policyManager, address _addressListRegistry)
        public
        AddressListRegistryPolicyBase(_policyManager, _addressListRegistry)
    {}

    // EXTERNAL FUNCTIONS

    /// notice Whether or not the policy can be disabled
    /// return canDisable_ True if the policy can be disabled
    function canDisable() external pure virtual override returns (bool canDisable_) {
        return true;
    }

    /// notice Provides a constant string identifier for a policy
    /// return identifier_ The identifer string
    function identifier() external pure override returns (string memory identifier_) {
        return "ALLOWED_DEPOSIT_RECIPIENTS";
    }

    /// notice Gets the implemented PolicyHooks for a policy
    /// return implementedHooks_ The implemented PolicyHooks
    function implementedHooks()
        external
        pure
        override
        returns (IPolicyManager.PolicyHook[] memory implementedHooks_)
    {
        implementedHooks_ = new IPolicyManager.PolicyHook[](1);
        implementedHooks_[0] = IPolicyManager.PolicyHook.PostBuyShares;

        return implementedHooks_;
    }

    /// notice Updates the policy settings for a fund
    /// param _comptrollerProxy The fund's ComptrollerProxy address
    /// param _encodedSettings Encoded settings to apply to a fund
    /// dev Used to assign a new list (not update items in that list)
    function updateFundSettings(address _comptrollerProxy, bytes calldata _encodedSettings)
        external
        override
        onlyPolicyManager
    {
        __updateListsForFund(_comptrollerProxy, _encodedSettings);
    }

    /// notice Apply the rule with the specified parameters of a PolicyHook
    /// param _comptrollerProxy The fund's ComptrollerProxy address
    /// param _encodedArgs Encoded args with which to validate the rule
    /// return isValid_ True if the rule passes
    /// dev onlyPolicyManager validation not necessary, as state is not updated and no events are fired
    function validateRule(
        address _comptrollerProxy,
        IPolicyManager.PolicyHook,
        bytes calldata _encodedArgs
    ) external override returns (bool isValid_) {
        (address buyer, , , ) = __decodePostBuySharesValidationData(_encodedArgs);

        return passesRule(_comptrollerProxy, buyer);
    }

    // PUBLIC FUNCTIONS

    /// notice Checks whether a particular condition passes the rule for a particular fund
    /// param _comptrollerProxy The fund's ComptrollerProxy address
    /// param _recipient The recipient of shares from the deposit
    /// return isValid_ True if the rule passes
    function passesRule(address _comptrollerProxy, address _recipient)
        public
        view
        returns (bool isValid_)
    {
        return
            AddressListRegistry(getAddressListRegistry()).isInSomeOfLists(
                getListIdsForFund(_comptrollerProxy),
                _recipient
            );
    }
}
