// SPDX-License-Identifier: GPL-3.0

/*
    This file is part of the Enzyme Protocol.

    (c) Enzyme Council <councilenzyme.finance>

    For the full license information, please view the LICENSE
    file that was distributed with this source code.
*/

pragma solidity ^0.6.12;

import "../../core/fund/comptroller/IComptroller.sol";
import "../../core/fund/vault/IVault.sol";

/// title PermissionedVaultActionMixin Contract
/// author Enzyme Council <securityenzyme.finance>
/// notice A mixin contract for extensions that can make permissioned vault calls
abstract contract PermissionedVaultActionMixin {
    /// notice Adds an external position to active external positions
    /// param _comptrollerProxy The ComptrollerProxy of the fund
    /// param _externalPosition The external position to be added
    function __addExternalPosition(address _comptrollerProxy, address _externalPosition) internal {
        IComptroller(_comptrollerProxy).permissionedVaultAction(
            IVault.VaultAction.AddExternalPosition,
            abi.encode(_externalPosition)
        );
    }

    /// notice Adds a tracked asset
    /// param _comptrollerProxy The ComptrollerProxy of the fund
    /// param _asset The asset to add
    function __addTrackedAsset(address _comptrollerProxy, address _asset) internal {
        IComptroller(_comptrollerProxy).permissionedVaultAction(
            IVault.VaultAction.AddTrackedAsset,
            abi.encode(_asset)
        );
    }

    /// notice Grants an allowance to a spender to use a fund's asset
    /// param _comptrollerProxy The ComptrollerProxy of the fund
    /// param _asset The asset for which to grant an allowance
    /// param _target The spender of the allowance
    /// param _amount The amount of the allowance
    function __approveAssetSpender(
        address _comptrollerProxy,
        address _asset,
        address _target,
        uint256 _amount
    ) internal {
        IComptroller(_comptrollerProxy).permissionedVaultAction(
            IVault.VaultAction.ApproveAssetSpender,
            abi.encode(_asset, _target, _amount)
        );
    }

    /// notice Burns fund shares for a particular account
    /// param _comptrollerProxy The ComptrollerProxy of the fund
    /// param _target The account for which to burn shares
    /// param _amount The amount of shares to burn
    function __burnShares(
        address _comptrollerProxy,
        address _target,
        uint256 _amount
    ) internal {
        IComptroller(_comptrollerProxy).permissionedVaultAction(
            IVault.VaultAction.BurnShares,
            abi.encode(_target, _amount)
        );
    }

    /// notice Executes a callOnExternalPosition
    /// param _comptrollerProxy The ComptrollerProxy of the fund
    /// param _data The encoded data for the call
    function __callOnExternalPosition(address _comptrollerProxy, bytes memory _data) internal {
        IComptroller(_comptrollerProxy).permissionedVaultAction(
            IVault.VaultAction.CallOnExternalPosition,
            _data
        );
    }

    /// notice Mints fund shares to a particular account
    /// param _comptrollerProxy The ComptrollerProxy of the fund
    /// param _target The account to which to mint shares
    /// param _amount The amount of shares to mint
    function __mintShares(
        address _comptrollerProxy,
        address _target,
        uint256 _amount
    ) internal {
        IComptroller(_comptrollerProxy).permissionedVaultAction(
            IVault.VaultAction.MintShares,
            abi.encode(_target, _amount)
        );
    }

    /// notice Removes an external position from the vaultProxy
    /// param _comptrollerProxy The ComptrollerProxy of the fund
    /// param _externalPosition The ExternalPosition to remove
    function __removeExternalPosition(address _comptrollerProxy, address _externalPosition)
        internal
    {
        IComptroller(_comptrollerProxy).permissionedVaultAction(
            IVault.VaultAction.RemoveExternalPosition,
            abi.encode(_externalPosition)
        );
    }

    /// notice Removes a tracked asset
    /// param _comptrollerProxy The ComptrollerProxy of the fund
    /// param _asset The asset to remove
    function __removeTrackedAsset(address _comptrollerProxy, address _asset) internal {
        IComptroller(_comptrollerProxy).permissionedVaultAction(
            IVault.VaultAction.RemoveTrackedAsset,
            abi.encode(_asset)
        );
    }

    /// notice Transfers fund shares from one account to another
    /// param _comptrollerProxy The ComptrollerProxy of the fund
    /// param _from The account from which to transfer shares
    /// param _to The account to which to transfer shares
    /// param _amount The amount of shares to transfer
    function __transferShares(
        address _comptrollerProxy,
        address _from,
        address _to,
        uint256 _amount
    ) internal {
        IComptroller(_comptrollerProxy).permissionedVaultAction(
            IVault.VaultAction.TransferShares,
            abi.encode(_from, _to, _amount)
        );
    }

    /// notice Withdraws an asset from the VaultProxy to a given account
    /// param _comptrollerProxy The ComptrollerProxy of the fund
    /// param _asset The asset to withdraw
    /// param _target The account to which to withdraw the asset
    /// param _amount The amount of asset to withdraw
    function __withdrawAssetTo(
        address _comptrollerProxy,
        address _asset,
        address _target,
        uint256 _amount
    ) internal {
        IComptroller(_comptrollerProxy).permissionedVaultAction(
            IVault.VaultAction.WithdrawAssetTo,
            abi.encode(_asset, _target, _amount)
        );
    }
}
