// SPDX-License-Identifier: MIT
pragma solidity ^0.8.13;

/**
 * title TokenTransferrerErrors
 */
interface TokenTransferrerErrors {
    /**
     * dev Revert with an error when an ERC721 transfer with amount other than
     *      one is attempted.
     *
     * param amount The amount of the ERC721 tokens to transfer.
     */
    error InvalidERC721TransferAmount(uint256 amount);

    /**
     * dev Revert with an error when attempting to fulfill an order where an
     *      item has an amount of zero.
     */
    error MissingItemAmount();

    /**
     * dev Revert with an error when attempting to fulfill an order where an
     *      item has unused parameters. This includes both the token and the
     *      identifier parameters for native transfers as well as the identifier
     *      parameter for ERC20 transfers. Note that the conduit does not
     *      perform this check, leaving it up to the calling channel to enforce
     *      when desired.
     */
    error UnusedItemParameters();

    /**
     * dev Revert with an error when an ERC20, ERC721, or ERC1155 token
     *      transfer reverts.
     *
     * param token      The token for which the transfer was attempted.
     * param from       The source of the attempted transfer.
     * param to         The recipient of the attempted transfer.
     * param identifier The identifier for the attempted transfer.
     * param amount     The amount for the attempted transfer.
     */
    error TokenTransferGenericFailure(
        address token,
        address from,
        address to,
        uint256 identifier,
        uint256 amount
    );

    /**
     * dev Revert with an error when a batch ERC1155 token transfer reverts.
     *
     * param token       The token for which the transfer was attempted.
     * param from        The source of the attempted transfer.
     * param to          The recipient of the attempted transfer.
     * param identifiers The identifiers for the attempted transfer.
     * param amounts     The amounts for the attempted transfer.
     */
    error ERC1155BatchTransferGenericFailure(
        address token,
        address from,
        address to,
        uint256[] identifiers,
        uint256[] amounts
    );

    /**
     * dev Revert with an error when an ERC20 token transfer returns a falsey
     *      value.
     *
     * param token      The token for which the ERC20 transfer was attempted.
     * param from       The source of the attempted ERC20 transfer.
     * param to         The recipient of the attempted ERC20 transfer.
     * param amount     The amount for the attempted ERC20 transfer.
     */
    error BadReturnValueFromERC20OnTransfer(
        address token,
        address from,
        address to,
        uint256 amount
    );

    /**
     * dev Revert with an error when an account being called as an assumed
     *      contract does not have code and returns no data.
     *
     * param account The account that should contain code.
     */
    error NoContract(address account);

    /**
     * dev Revert with an error when attempting to execute an 1155 batch
     *      transfer using calldata not produced by default ABI encoding or with
     *      different lengths for ids and amounts arrays.
     */
    error Invalid1155BatchTransferEncoding();
}
