// SPDX-License-Identifier: MIT
// OpenZeppelin Contracts v4.4.1 (utils/Context.sol)

pragma solidity ^0.8.0;

/**
 * @dev Provides information about the current execution context, including the
 * sender of the transaction and its data. While these are generally available
 * via msg.sender and msg.data, they should not be accessed in such a direct
 * manner, since when dealing with meta-transactions the account sending and
 * paying for execution may not be the actual sender (as far as an application
 * is concerned).
 *
 * This contract is only required for intermediate, library-like contracts.
 */
abstract contract Context {
    function _msgSender() internal view virtual returns (address) {
        return msg.sender;
    }

    function _msgData() internal view virtual returns (bytes calldata) {
        return msg.data;
    }
}


// OpenZeppelin Contracts (last updated v4.7.0) (access/Ownable.sol)

//pragma solidity ^0.8.0;

/**
 * @dev Contract module which provides a basic access control mechanism, where
 * there is an account (an owner) that can be granted exclusive access to
 * specific functions.
 *
 * By default, the owner account will be the one that deploys the contract. This
 * can later be changed with {transferOwnership}.
 *
 * This module is used through inheritance. It will make available the modifier
 * `onlyOwner`, which can be applied to your functions to restrict their use to
 * the owner.
 */
abstract contract Ownable is Context {
    address private _owner;

    event OwnershipTransferred(address indexed previousOwner, address indexed newOwner);

    /**
     * @dev Initializes the contract setting the deployer as the initial owner.
     */
    constructor() {
        _transferOwnership(_msgSender());
    }

    /**
     * @dev Throws if called by any account other than the owner.
     */
    modifier onlyOwner() {
        _checkOwner();
        _;
    }

    /**
     * @dev Returns the address of the current owner.
     */
    function owner() public view virtual returns (address) {
        return _owner;
    }

    /**
     * @dev Throws if the sender is not the owner.
     */
    function _checkOwner() internal view virtual {
        require(owner() == _msgSender(), "Ownable: caller is not the owner");
    }

    /**
     * @dev Leaves the contract without owner. It will not be possible to call
     * `onlyOwner` functions anymore. Can only be called by the current owner.
     *
     * NOTE: Renouncing ownership will leave the contract without an owner,
     * thereby removing any functionality that is only available to the owner.
     */
    function renounceOwnership() public virtual onlyOwner {
        _transferOwnership(address(0));
    }

    /**
     * @dev Transfers ownership of the contract to a new account (`newOwner`).
     * Can only be called by the current owner.
     */
    function transferOwnership(address newOwner) public virtual onlyOwner {
        require(newOwner != address(0), "Ownable: new owner is the zero address");
        _transferOwnership(newOwner);
    }

    /**
     * @dev Transfers ownership of the contract to a new account (`newOwner`).
     * Internal function without access restriction.
     */
    function _transferOwnership(address newOwner) internal virtual {
        address oldOwner = _owner;
        _owner = newOwner;
        emit OwnershipTransferred(oldOwner, newOwner);
    }
}


// File contracts/Confirmer.sol
//pragma solidity >=0.8.0 <0.9.0;

abstract contract Confirmer is Ownable {

    event AddConfirmer (address account);
    event RemoveConfirmer (address account);
    event Confirmed (address account, uint256 confirmedTime);
    
    struct ConfirmInfo {
        bool confirmed;
        uint256 validTime;
    }
    
    mapping(address => ConfirmInfo) private _confirmInfos;

    address[] public _confirmers;

    constructor() {
    }

    modifier isConfirmed() {
        require(_checkConfirmed(), "Insufficient execution conditions.");
        _;
    }

    modifier isConfirmer(address walletAddress) {
        require(_checkConfirmer(walletAddress), "This address or caller is not the confirmer.");
        _;
    }

    function _checkConfirmed() 
        internal
        view
        returns (bool)
    {
        uint8 confirmedCount = 0;
        for (uint8 i = 0; i < _confirmers.length; i++) {
            if (_confirmInfos[_confirmers[i]].confirmed == true 
                && _confirmInfos[_confirmers[i]].validTime > block.timestamp) 
            {
                confirmedCount++;
            }
        }

        if (confirmedCount > 1) return true;
        else return false;
    }

    function _checkConfirmer(address walletAddress)
        internal
        view
        returns (bool)
    {
        for (uint8 i = 0; i < _confirmers.length; i++) {
            if (_confirmers[i] == walletAddress) {
                return true;
            }
        }

        return false;
    }    

    function addConfirmer(address walletAddress) 
        public
        onlyOwner
        isConfirmed
    {
        require(!_checkConfirmer(walletAddress), "Aleady exist confirmer.");

        _confirmers.push(walletAddress);
        _resetConfirmed();

        emit AddConfirmer(walletAddress);
    }

    function removeConfirmer(address walletAddress)
        public
        onlyOwner
        isConfirmed
        isConfirmer(walletAddress)
    {
        require(_confirmers.length > 3, "Must be at least 3 confirmers.");
        uint8 j = 0;
        for (uint8 i = 0; i < _confirmers.length; i++) {
            if (_confirmers[i] != walletAddress) {
                _confirmers[j] = _confirmers[i];
                j++;
            } else if (_confirmers[i] == walletAddress){
                delete _confirmers[i];
                delete _confirmInfos[_confirmers[i]];
            }
        }

        _confirmers.pop();
        _resetConfirmed();

        emit RemoveConfirmer (walletAddress);
    }

    function _resetConfirmed()
        internal
    {
        require((_checkConfirmer(msg.sender) || msg.sender == owner()), "Caller is not the owner or confirmer.");
        for (uint8 i = 0; i < _confirmers.length; i++) {
            _confirmInfos[_confirmers[i]].confirmed = false;
            _confirmInfos[_confirmers[i]].validTime = block.timestamp;
        }     
    }

    function toConfirm()
        public
        isConfirmer(msg.sender)
    {
        _confirmInfos[msg.sender].confirmed = true;
        _confirmInfos[msg.sender].validTime = block.timestamp + 86400;

        emit Confirmed(msg.sender, block.timestamp);
    }

    function getConfirmer() 
        public
        view
        onlyOwner
        returns (address[] memory) 
    {
        return (_confirmers);
    }

    function getConfirmed(address walletAddress) 
        public
        view
        onlyOwner
        returns (bool, uint256)
    {
        return (_confirmInfos[walletAddress].confirmed, _confirmInfos[walletAddress].validTime);
    }    

}


// OpenZeppelin Contracts (last updated v4.7.0) (security/Pausable.sol)

//pragma solidity ^0.8.0;

/**
 * @dev Contract module which allows children to implement an emergency stop
 * mechanism that can be triggered by an authorized account.
 *
 * This module is used through inheritance. It will make available the
 * modifiers `whenNotPaused` and `whenPaused`, which can be applied to
 * the functions of your contract. Note that they will not be pausable by
 * simply including this module, only once the modifiers are put in place.
 */
abstract contract Pausable is Context {
    /**
     * @dev Emitted when the pause is triggered by `account`.
     */
    event Paused(address account);

    /**
     * @dev Emitted when the pause is lifted by `account`.
     */
    event Unpaused(address account);

    bool private _paused;

    /**
     * @dev Initializes the contract in unpaused state.
     */
    constructor() {
        _paused = false;
    }

    /**
     * @dev Modifier to make a function callable only when the contract is not paused.
     *
     * Requirements:
     *
     * - The contract must not be paused.
     */
    modifier whenNotPaused() {
        _requireNotPaused();
        _;
    }

    /**
     * @dev Modifier to make a function callable only when the contract is paused.
     *
     * Requirements:
     *
     * - The contract must be paused.
     */
    modifier whenPaused() {
        _requirePaused();
        _;
    }

    /**
     * @dev Returns true if the contract is paused, and false otherwise.
     */
    function paused() public view virtual returns (bool) {
        return _paused;
    }

    /**
     * @dev Throws if the contract is paused.
     */
    function _requireNotPaused() internal view virtual {
        require(!paused(), "Pausable: paused");
    }

    /**
     * @dev Throws if the contract is not paused.
     */
    function _requirePaused() internal view virtual {
        require(paused(), "Pausable: not paused");
    }

    /**
     * @dev Triggers stopped state.
     *
     * Requirements:
     *
     * - The contract must not be paused.
     */
    function _pause() internal virtual whenNotPaused {
        _paused = true;
        emit Paused(_msgSender());
    }

    /**
     * @dev Returns to normal state.
     *
     * Requirements:
     *
     * - The contract must be paused.
     */
    function _unpause() internal virtual whenPaused {
        _paused = false;
        emit Unpaused(_msgSender());
    }
}


// OpenZeppelin Contracts (last updated v4.6.0) (token/ERC20/IERC20.sol)

//pragma solidity ^0.8.0;

/**
 * @dev Interface of the ERC20 standard as defined in the EIP.
 */
interface IERC20 {
    /**
     * @dev Emitted when `value` tokens are moved from one account (`from`) to
     * another (`to`).
     *
     * Note that `value` may be zero.
     */
    event Transfer(address indexed from, address indexed to, uint256 value);

    /**
     * @dev Emitted when the allowance of a `spender` for an `owner` is set by
     * a call to {approve}. `value` is the new allowance.
     */
    event Approval(address indexed owner, address indexed spender, uint256 value);

    /**
     * @dev Returns the amount of tokens in existence.
     */
    function totalSupply() external view returns (uint256);

    /**
     * @dev Returns the amount of tokens owned by `account`.
     */
    function balanceOf(address account) external view returns (uint256);

    /**
     * @dev Moves `amount` tokens from the caller's account to `to`.
     *
     * Returns a boolean value indicating whether the operation succeeded.
     *
     * Emits a {Transfer} event.
     */
    function transfer(address to, uint256 amount) external returns (bool);

    /**
     * @dev Returns the remaining number of tokens that `spender` will be
     * allowed to spend on behalf of `owner` through {transferFrom}. This is
     * zero by default.
     *
     * This value changes when {approve} or {transferFrom} are called.
     */
    function allowance(address owner, address spender) external view returns (uint256);

    /**
     * @dev Sets `amount` as the allowance of `spender` over the caller's tokens.
     *
     * Returns a boolean value indicating whether the operation succeeded.
     *
     * IMPORTANT: Beware that changing an allowance with this method brings the risk
     * that someone may use both the old and the new allowance by unfortunate
     * transaction ordering. One possible solution to mitigate this race
     * condition is to first reduce the spender's allowance to 0 and set the
     * desired value afterwards:
     * https://github.com/ethereum/EIPs/issues/20#issuecomment-263524729
     *
     * Emits an {Approval} event.
     */
    function approve(address spender, uint256 amount) external returns (bool);

    /**
     * @dev Moves `amount` tokens from `from` to `to` using the
     * allowance mechanism. `amount` is then deducted from the caller's
     * allowance.
     *
     * Returns a boolean value indicating whether the operation succeeded.
     *
     * Emits a {Transfer} event.
     */
    function transferFrom(
        address from,
        address to,
        uint256 amount
    ) external returns (bool);
}


// OpenZeppelin Contracts v4.4.1 (token/ERC20/extensions/IERC20Metadata.sol)

//pragma solidity ^0.8.0;

/**
 * @dev Interface for the optional metadata functions from the ERC20 standard.
 *
 * _Available since v4.1._
 */
interface IERC20Metadata is IERC20 {
    /**
     * @dev Returns the name of the token.
     */
    function name() external view returns (string memory);

    /**
     * @dev Returns the symbol of the token.
     */
    function symbol() external view returns (string memory);

    /**
     * @dev Returns the decimals places of the token.
     */
    function decimals() external view returns (uint8);
}


// OpenZeppelin Contracts (last updated v4.7.0) (token/ERC20/ERC20.sol)

//pragma solidity ^0.8.0;



/**
 * @dev Implementation of the {IERC20} interface.
 *
 * This implementation is agnostic to the way tokens are created. This means
 * that a supply mechanism has to be added in a derived contract using {_mint}.
 * For a generic mechanism see {ERC20PresetMinterPauser}.
 *
 * TIP: For a detailed writeup see our guide
 * https://forum.zeppelin.solutions/t/how-to-implement-erc20-supply-mechanisms/226[How
 * to implement supply mechanisms].
 *
 * We have followed general OpenZeppelin Contracts guidelines: functions revert
 * instead returning `false` on failure. This behavior is nonetheless
 * conventional and does not conflict with the expectations of ERC20
 * applications.
 *
 * Additionally, an {Approval} event is emitted on calls to {transferFrom}.
 * This allows applications to reconstruct the allowance for all accounts just
 * by listening to said events. Other implementations of the EIP may not emit
 * these events, as it isn't required by the specification.
 *
 * Finally, the non-standard {decreaseAllowance} and {increaseAllowance}
 * functions have been added to mitigate the well-known issues around setting
 * allowances. See {IERC20-approve}.
 */
contract ERC20 is Context, IERC20, IERC20Metadata {
    mapping(address => uint256) private _balances;

    mapping(address => mapping(address => uint256)) private _allowances;

    uint256 private _totalSupply;

    string private _name;
    string private _symbol;

    /**
     * @dev Sets the values for {name} and {symbol}.
     *
     * The default value of {decimals} is 18. To select a different value for
     * {decimals} you should overload it.
     *
     * All two of these values are immutable: they can only be set once during
     * construction.
     */
    constructor(string memory name_, string memory symbol_) {
        _name = name_;
        _symbol = symbol_;
    }

    /**
     * @dev Returns the name of the token.
     */
    function name() public view virtual override returns (string memory) {
        return _name;
    }

    /**
     * @dev Returns the symbol of the token, usually a shorter version of the
     * name.
     */
    function symbol() public view virtual override returns (string memory) {
        return _symbol;
    }

    /**
     * @dev Returns the number of decimals used to get its user representation.
     * For example, if `decimals` equals `2`, a balance of `505` tokens should
     * be displayed to a user as `5.05` (`505 / 10 ** 2`).
     *
     * Tokens usually opt for a value of 18, imitating the relationship between
     * Ether and Wei. This is the value {ERC20} uses, unless this function is
     * overridden;
     *
     * NOTE: This information is only used for _display_ purposes: it in
     * no way affects any of the arithmetic of the contract, including
     * {IERC20-balanceOf} and {IERC20-transfer}.
     */
    function decimals() public view virtual override returns (uint8) {
        return 18;
    }

    /**
     * @dev See {IERC20-totalSupply}.
     */
    function totalSupply() public view virtual override returns (uint256) {
        return _totalSupply;
    }

    /**
     * @dev See {IERC20-balanceOf}.
     */
    function balanceOf(address account) public view virtual override returns (uint256) {
        return _balances[account];
    }

    /**
     * @dev See {IERC20-transfer}.
     *
     * Requirements:
     *
     * - `to` cannot be the zero address.
     * - the caller must have a balance of at least `amount`.
     */
    function transfer(address to, uint256 amount) public virtual override returns (bool) {
        address owner = _msgSender();
        _transfer(owner, to, amount);
        return true;
    }

    /**
     * @dev See {IERC20-allowance}.
     */
    function allowance(address owner, address spender) public view virtual override returns (uint256) {
        return _allowances[owner][spender];
    }

    /**
     * @dev See {IERC20-approve}.
     *
     * NOTE: If `amount` is the maximum `uint256`, the allowance is not updated on
     * `transferFrom`. This is semantically equivalent to an infinite approval.
     *
     * Requirements:
     *
     * - `spender` cannot be the zero address.
     */
    function approve(address spender, uint256 amount) public virtual override returns (bool) {
        address owner = _msgSender();
        _approve(owner, spender, amount);
        return true;
    }

    /**
     * @dev See {IERC20-transferFrom}.
     *
     * Emits an {Approval} event indicating the updated allowance. This is not
     * required by the EIP. See the note at the beginning of {ERC20}.
     *
     * NOTE: Does not update the allowance if the current allowance
     * is the maximum `uint256`.
     *
     * Requirements:
     *
     * - `from` and `to` cannot be the zero address.
     * - `from` must have a balance of at least `amount`.
     * - the caller must have allowance for ``from``'s tokens of at least
     * `amount`.
     */
    function transferFrom(
        address from,
        address to,
        uint256 amount
    ) public virtual override returns (bool) {
        address spender = _msgSender();
        _spendAllowance(from, spender, amount);
        _transfer(from, to, amount);
        return true;
    }

    /**
     * @dev Atomically increases the allowance granted to `spender` by the caller.
     *
     * This is an alternative to {approve} that can be used as a mitigation for
     * problems described in {IERC20-approve}.
     *
     * Emits an {Approval} event indicating the updated allowance.
     *
     * Requirements:
     *
     * - `spender` cannot be the zero address.
     */
    function increaseAllowance(address spender, uint256 addedValue) public virtual returns (bool) {
        address owner = _msgSender();
        _approve(owner, spender, allowance(owner, spender) + addedValue);
        return true;
    }

    /**
     * @dev Atomically decreases the allowance granted to `spender` by the caller.
     *
     * This is an alternative to {approve} that can be used as a mitigation for
     * problems described in {IERC20-approve}.
     *
     * Emits an {Approval} event indicating the updated allowance.
     *
     * Requirements:
     *
     * - `spender` cannot be the zero address.
     * - `spender` must have allowance for the caller of at least
     * `subtractedValue`.
     */
    function decreaseAllowance(address spender, uint256 subtractedValue) public virtual returns (bool) {
        address owner = _msgSender();
        uint256 currentAllowance = allowance(owner, spender);
        require(currentAllowance >= subtractedValue, "ERC20: decreased allowance below zero");
        unchecked {
            _approve(owner, spender, currentAllowance - subtractedValue);
        }

        return true;
    }

    /**
     * @dev Moves `amount` of tokens from `from` to `to`.
     *
     * This internal function is equivalent to {transfer}, and can be used to
     * e.g. implement automatic token fees, slashing mechanisms, etc.
     *
     * Emits a {Transfer} event.
     *
     * Requirements:
     *
     * - `from` cannot be the zero address.
     * - `to` cannot be the zero address.
     * - `from` must have a balance of at least `amount`.
     */
    function _transfer(
        address from,
        address to,
        uint256 amount
    ) internal virtual {
        require(from != address(0), "ERC20: transfer from the zero address");
        require(to != address(0), "ERC20: transfer to the zero address");

        _beforeTokenTransfer(from, to, amount);

        uint256 fromBalance = _balances[from];
        require(fromBalance >= amount, "ERC20: transfer amount exceeds balance");
        unchecked {
            _balances[from] = fromBalance - amount;
        }
        _balances[to] += amount;

        emit Transfer(from, to, amount);

        _afterTokenTransfer(from, to, amount);
    }

    /** @dev Creates `amount` tokens and assigns them to `account`, increasing
     * the total supply.
     *
     * Emits a {Transfer} event with `from` set to the zero address.
     *
     * Requirements:
     *
     * - `account` cannot be the zero address.
     */
    function _mint(address account, uint256 amount) internal virtual {
        require(account != address(0), "ERC20: mint to the zero address");

        _beforeTokenTransfer(address(0), account, amount);

        _totalSupply += amount;
        _balances[account] += amount;
        emit Transfer(address(0), account, amount);

        _afterTokenTransfer(address(0), account, amount);
    }

    /**
     * @dev Destroys `amount` tokens from `account`, reducing the
     * total supply.
     *
     * Emits a {Transfer} event with `to` set to the zero address.
     *
     * Requirements:
     *
     * - `account` cannot be the zero address.
     * - `account` must have at least `amount` tokens.
     */
    function _burn(address account, uint256 amount) internal virtual {
        require(account != address(0), "ERC20: burn from the zero address");

        _beforeTokenTransfer(account, address(0), amount);

        uint256 accountBalance = _balances[account];
        require(accountBalance >= amount, "ERC20: burn amount exceeds balance");
        unchecked {
            _balances[account] = accountBalance - amount;
        }
        _totalSupply -= amount;

        emit Transfer(account, address(0), amount);

        _afterTokenTransfer(account, address(0), amount);
    }

    /**
     * @dev Sets `amount` as the allowance of `spender` over the `owner` s tokens.
     *
     * This internal function is equivalent to `approve`, and can be used to
     * e.g. set automatic allowances for certain subsystems, etc.
     *
     * Emits an {Approval} event.
     *
     * Requirements:
     *
     * - `owner` cannot be the zero address.
     * - `spender` cannot be the zero address.
     */
    function _approve(
        address owner,
        address spender,
        uint256 amount
    ) internal virtual {
        require(owner != address(0), "ERC20: approve from the zero address");
        require(spender != address(0), "ERC20: approve to the zero address");

        _allowances[owner][spender] = amount;
        emit Approval(owner, spender, amount);
    }

    /**
     * @dev Updates `owner` s allowance for `spender` based on spent `amount`.
     *
     * Does not update the allowance amount in case of infinite allowance.
     * Revert if not enough allowance is available.
     *
     * Might emit an {Approval} event.
     */
    function _spendAllowance(
        address owner,
        address spender,
        uint256 amount
    ) internal virtual {
        uint256 currentAllowance = allowance(owner, spender);
        if (currentAllowance != type(uint256).max) {
            require(currentAllowance >= amount, "ERC20: insufficient allowance");
            unchecked {
                _approve(owner, spender, currentAllowance - amount);
            }
        }
    }

    /**
     * @dev Hook that is called before any transfer of tokens. This includes
     * minting and burning.
     *
     * Calling conditions:
     *
     * - when `from` and `to` are both non-zero, `amount` of ``from``'s tokens
     * will be transferred to `to`.
     * - when `from` is zero, `amount` tokens will be minted for `to`.
     * - when `to` is zero, `amount` of ``from``'s tokens will be burned.
     * - `from` and `to` are never both zero.
     *
     * To learn more about hooks, head to xref:ROOT:extending-contracts.adoc#using-hooks[Using Hooks].
     */
    function _beforeTokenTransfer(
        address from,
        address to,
        uint256 amount
    ) internal virtual {}

    /**
     * @dev Hook that is called after any transfer of tokens. This includes
     * minting and burning.
     *
     * Calling conditions:
     *
     * - when `from` and `to` are both non-zero, `amount` of ``from``'s tokens
     * has been transferred to `to`.
     * - when `from` is zero, `amount` tokens have been minted for `to`.
     * - when `to` is zero, `amount` of ``from``'s tokens have been burned.
     * - `from` and `to` are never both zero.
     *
     * To learn more about hooks, head to xref:ROOT:extending-contracts.adoc#using-hooks[Using Hooks].
     */
    function _afterTokenTransfer(
        address from,
        address to,
        uint256 amount
    ) internal virtual {}
}


// OpenZeppelin Contracts (last updated v4.5.0) (token/ERC20/extensions/ERC20Burnable.sol)

//pragma solidity ^0.8.0;


/**
 * @dev Extension of {ERC20} that allows token holders to destroy both their own
 * tokens and those that they have an allowance for, in a way that can be
 * recognized off-chain (via event analysis).
 */
abstract contract ERC20Burnable is Context, ERC20 {
    /**
     * @dev Destroys `amount` tokens from the caller.
     *
     * See {ERC20-_burn}.
     */
    function burn(uint256 amount) public virtual {
        _burn(_msgSender(), amount);
    }

    /**
     * @dev Destroys `amount` tokens from `account`, deducting from the caller's
     * allowance.
     *
     * See {ERC20-_burn} and {ERC20-allowance}.
     *
     * Requirements:
     *
     * - the caller must have allowance for ``accounts``'s tokens of at least
     * `amount`.
     */
    function burnFrom(address account, uint256 amount) public virtual {
        _spendAllowance(account, _msgSender(), amount);
        _burn(account, amount);
    }
}



//pragma solidity >=0.8.0 <0.9.0;


contract CONUTToken is ERC20, ERC20Burnable, Pausable, Ownable, Confirmer {

    event TimeLock(address indexed account, uint256 amount, uint256 startTime, uint256 releaseMonths);
    event TimeLockUpdate (address indexed account, uint256 valueIndex, uint256 value);
    
    struct LockInfo {
        uint256 amount;    
        uint256 startTime;
        uint256 releaseMonths;
    }
    
    mapping(address => LockInfo) private _lockInfos;

    address[] private _lockedWallets;

    constructor() ERC20("CONUT", "CONUT") {
        _mint(msg.sender, 10000000000 * 10 ** decimals());

        _confirmers.push(msg.sender);
        _resetConfirmed();
    }

    function pause() public onlyOwner isConfirmed 
    {
        _pause();
        _resetConfirmed();
    }

    function unpause() public onlyOwner isConfirmed 
    {
        _unpause();
        _resetConfirmed();
    }

    function transferOwnership(address newOwner) public isConfirmer(msg.sender) isConfirmed override
    {
        require(newOwner != address(0), "Ownable: new owner is the zero address");

        super._transferOwnership(newOwner);
        _resetConfirmed();
    }

    function renounceOwnership() public onlyOwner isConfirmed override
    {
        super.renounceOwnership();
        _resetConfirmed();
    }

    function setLock(address walletAddress, uint256 startTime, uint256 releaseMonths, uint256 amount) 
        public
        onlyOwner 
    {
        require(_lockInfos[walletAddress].amount < 1, "Aleady exist lock info.");
        require(block.timestamp < startTime, "ERC20: Current time is greater than start time.");
        require(releaseMonths > 0, "ERC20: ReleaseMonths is greater than 0.");
        require(amount > 0, "ERC20: Amount is greater than 0.");
        
        _lockInfos[walletAddress] = LockInfo(amount, startTime, releaseMonths);
        _lockedWallets.push(walletAddress);

        emit TimeLock( walletAddress, amount, startTime, releaseMonths ); 
    }

    function setLockReleaseMonths(address walletAddress, uint256 releaseMonths)
        public
        onlyOwner
        isConfirmed
    {
        require(_lockInfos[walletAddress].amount > 0, "Not exist lock info.");
        require(releaseMonths > 0, "ERC20: ReleaseMonths is greater than 0.");
        _lockInfos[walletAddress].releaseMonths = releaseMonths;

        emit TimeLockUpdate (walletAddress, 3, releaseMonths);
        _resetConfirmed();
    }

    function setLockStartTime(address walletAddress, uint256 startTime)
        public
        onlyOwner
        isConfirmed
    {
        require(_lockInfos[walletAddress].amount > 0, "Not exist lock info.");
        require(block.timestamp < startTime, "ERC20: Current time is greater than start time");
        _lockInfos[walletAddress].startTime = startTime;

        emit TimeLockUpdate (walletAddress, 2, startTime);
        _resetConfirmed();
    }

    function setLockAmount(address walletAddress, uint256 amount)
        public
        onlyOwner
        isConfirmed
    {
        require(_lockInfos[walletAddress].amount > 0, "Not exist lock info.");
        require(amount > 0, "ERC20: Amount is greater than 0.");
        _lockInfos[walletAddress].amount = amount;

        emit TimeLockUpdate (walletAddress, 1, amount);
        _resetConfirmed();
    }

    function getLockInfo(address walletAddress) 
        public 
        view 
        returns (uint256 lockAmount, uint256 startTime, uint256 releaseMonths, uint256 released) 
    {
        require(msg.sender == walletAddress || msg.sender == owner(), "Caller is not the owner.");
        require(_lockInfos[walletAddress].amount > 0, "Not exist lock info.");
        
        uint256 unLockAmount = _getUnLockAmount(walletAddress, block.timestamp);

        return (_lockInfos[walletAddress].amount, _lockInfos[walletAddress].startTime, _lockInfos[walletAddress].releaseMonths, unLockAmount);
    }

    function getLockWallets()
        public
        view
        virtual
        onlyOwner
        returns (address[] memory) 
    {
        return (_lockedWallets);
    }

    function _getUnLockAmount(address walletAddress, uint256 timestamp)
        internal
        view
        returns (uint256 unLockAmount)
    {
        uint256 lockAmount = _lockInfos[walletAddress].amount;
        uint256 lockStartTime = _lockInfos[walletAddress].startTime;
        uint256 releaseMonths = _lockInfos[walletAddress].releaseMonths;

        if (timestamp < lockStartTime) return 0;

        uint256 checkReleasedSecond = timestamp - lockStartTime;
        uint256 checkReleasedMonth = checkReleasedSecond / (86400 * 30);  //per 30day        

        if (releaseMonths <= checkReleasedMonth) {
            return lockAmount;
        } else if (checkReleasedMonth < 1) {
            return 0;
        } else {
            return lockAmount * checkReleasedMonth / releaseMonths;
        }
    }

    function _isLocked(address walletAddress, uint256 amount) 
        internal 
        view 
        returns (bool) 
    {
        if (_lockInfos[walletAddress].amount != 0) {
            uint256 unLockAmount = _getUnLockAmount(walletAddress, block.timestamp);
            return (balanceOf(walletAddress) - (_lockInfos[walletAddress].amount - unLockAmount)) < amount;
        } else {
            return false;
        }
    }

    function _beforeTokenTransfer(address from, address to, uint256 amount)
        internal
        whenNotPaused
        override
    {
        require( !_isLocked( from, amount ) , "ERC20: Locked balance.");
        super._beforeTokenTransfer(from, to, amount);
    }

}