// SPDX-License-Identifier: MIT
pragma solidity ^0.8.0;

import { Payment, BidState } from "../TellerV2Storage.sol";
import { Collateral } from "./escrow/ICollateralEscrowV1.sol";

interface ITellerV2 {
    /**
     * notice Function for a borrower to create a bid for a loan.
     * param _lendingToken The lending token asset requested to be borrowed.
     * param _marketplaceId The unique id of the marketplace for the bid.
     * param _principal The principal amount of the loan bid.
     * param _duration The recurrent length of time before which a payment is due.
     * param _APR The proposed interest rate for the loan bid.
     * param _metadataURI The URI for additional borrower loan information as part of loan bid.
     * param _receiver The address where the loan amount will be sent to.
     */
    function submitBid(
        address _lendingToken,
        uint256 _marketplaceId,
        uint256 _principal,
        uint32 _duration,
        uint16 _APR,
        string calldata _metadataURI,
        address _receiver
    ) external returns (uint256 bidId_);

    /**
     * notice Function for a borrower to create a bid for a loan with Collateral.
     * param _lendingToken The lending token asset requested to be borrowed.
     * param _marketplaceId The unique id of the marketplace for the bid.
     * param _principal The principal amount of the loan bid.
     * param _duration The recurrent length of time before which a payment is due.
     * param _APR The proposed interest rate for the loan bid.
     * param _metadataURI The URI for additional borrower loan information as part of loan bid.
     * param _receiver The address where the loan amount will be sent to.
     * param _collateralInfo Additional information about the collateral asset.
     */
    function submitBid(
        address _lendingToken,
        uint256 _marketplaceId,
        uint256 _principal,
        uint32 _duration,
        uint16 _APR,
        string calldata _metadataURI,
        address _receiver,
        Collateral[] calldata _collateralInfo
    ) external returns (uint256 bidId_);

    /**
     * notice Function for a lender to accept a proposed loan bid.
     * param _bidId The id of the loan bid to accept.
     */
    function lenderAcceptBid(uint256 _bidId)
        external
        returns (
            uint256 amountToProtocol,
            uint256 amountToMarketplace,
            uint256 amountToBorrower
        );

    function calculateAmountDue(uint256 _bidId)
        external
        view
        returns (Payment memory due);

    /**
     * notice Function for users to make the minimum amount due for an active loan.
     * param _bidId The id of the loan to make the payment towards.
     */
    function repayLoanMinimum(uint256 _bidId) external;

    /**
     * notice Function for users to repay an active loan in full.
     * param _bidId The id of the loan to make the payment towards.
     */
    function repayLoanFull(uint256 _bidId) external;

    /**
     * notice Function for users to make a payment towards an active loan.
     * param _bidId The id of the loan to make the payment towards.
     * param _amount The amount of the payment.
     */
    function repayLoan(uint256 _bidId, uint256 _amount) external;

    /**
     * notice Checks to see if a borrower is delinquent.
     * param _bidId The id of the loan bid to check for.
     */
    function isLoanDefaulted(uint256 _bidId) external view returns (bool);

    /**
     * notice Checks to see if a loan was delinquent for longer than liquidation delay.
     * param _bidId The id of the loan bid to check for.
     */
    function isLoanLiquidateable(uint256 _bidId) external view returns (bool);

    /**
     * notice Checks to see if a borrower is delinquent.
     * param _bidId The id of the loan bid to check for.
     */
    function isPaymentLate(uint256 _bidId) external view returns (bool);

    function getBidState(uint256 _bidId) external view returns (BidState);

    function getBorrowerActiveLoanIds(address _borrower)
        external
        view
        returns (uint256[] memory);

    /**
     * notice Returns the borrower address for a given bid.
     * param _bidId The id of the bid/loan to get the borrower for.
     * return borrower_ The address of the borrower associated with the bid.
     */
    function getLoanBorrower(uint256 _bidId)
        external
        view
        returns (address borrower_);

    /**
     * notice Returns the lender address for a given bid.
     * param _bidId The id of the bid/loan to get the lender for.
     * return lender_ The address of the lender associated with the bid.
     */
    function getLoanLender(uint256 _bidId)
        external
        view
        returns (address lender_);

    function getLoanLendingToken(uint256 _bidId)
        external
        view
        returns (address token_);

    function getLoanMarketId(uint256 _bidId) external view returns (uint256);

    function getLoanSummary(uint256 _bidId)
        external
        view
        returns (
            address borrower,
            address lender,
            uint256 marketId,
            address principalTokenAddress,
            uint256 principalAmount,
            uint32 acceptedTimestamp,
            uint32 lastRepaidTimestamp,
            BidState bidState
        );
}
