pragma solidity >=0.8.0 <0.9.0;
// SPDX-License-Identifier: MIT

import "./IASRegistry.sol";
import "./IEASEIP712Verifier.sol";

/**
 * title EAS - Ethereum Attestation Service interface
 */
interface IEAS {
    /**
     * dev A struct representing a single attestation.
     */
    struct Attestation {
        // A unique identifier of the attestation.
        bytes32 uuid;
        // A unique identifier of the AS.
        bytes32 schema;
        // The recipient of the attestation.
        address recipient;
        // The attester/sender of the attestation.
        address attester;
        // The time when the attestation was created (Unix timestamp).
        uint256 time;
        // The time when the attestation expires (Unix timestamp).
        uint256 expirationTime;
        // The time when the attestation was revoked (Unix timestamp).
        uint256 revocationTime;
        // The UUID of the related attestation.
        bytes32 refUUID;
        // Custom attestation data.
        bytes data;
    }

    /**
     * dev Triggered when an attestation has been made.
     *
     * param recipient The recipient of the attestation.
     * param attester The attesting account.
     * param uuid The UUID the revoked attestation.
     * param schema The UUID of the AS.
     */
    event Attested(
        address indexed recipient,
        address indexed attester,
        bytes32 uuid,
        bytes32 indexed schema
    );

    /**
     * dev Triggered when an attestation has been revoked.
     *
     * param recipient The recipient of the attestation.
     * param attester The attesting account.
     * param schema The UUID of the AS.
     * param uuid The UUID the revoked attestation.
     */
    event Revoked(
        address indexed recipient,
        address indexed attester,
        bytes32 uuid,
        bytes32 indexed schema
    );

    /**
     * dev Returns the address of the AS global registry.
     *
     * return The address of the AS global registry.
     */
    function getASRegistry() external view returns (IASRegistry);

    /**
     * dev Returns the address of the EIP712 verifier used to verify signed attestations.
     *
     * return The address of the EIP712 verifier used to verify signed attestations.
     */
    function getEIP712Verifier() external view returns (IEASEIP712Verifier);

    /**
     * dev Returns the global counter for the total number of attestations.
     *
     * return The global counter for the total number of attestations.
     */
    function getAttestationsCount() external view returns (uint256);

    /**
     * dev Attests to a specific AS.
     *
     * param recipient The recipient of the attestation.
     * param schema The UUID of the AS.
     * param expirationTime The expiration time of the attestation.
     * param refUUID An optional related attestation's UUID.
     * param data Additional custom data.
     *
     * return The UUID of the new attestation.
     */
    function attest(
        address recipient,
        bytes32 schema,
        uint256 expirationTime,
        bytes32 refUUID,
        bytes calldata data
    ) external payable returns (bytes32);

    /**
     * dev Attests to a specific AS using a provided EIP712 signature.
     *
     * param recipient The recipient of the attestation.
     * param schema The UUID of the AS.
     * param expirationTime The expiration time of the attestation.
     * param refUUID An optional related attestation's UUID.
     * param data Additional custom data.
     * param attester The attesting account.
     * param v The recovery ID.
     * param r The x-coordinate of the nonce R.
     * param s The signature data.
     *
     * return The UUID of the new attestation.
     */
    function attestByDelegation(
        address recipient,
        bytes32 schema,
        uint256 expirationTime,
        bytes32 refUUID,
        bytes calldata data,
        address attester,
        uint8 v,
        bytes32 r,
        bytes32 s
    ) external payable returns (bytes32);

    /**
     * dev Revokes an existing attestation to a specific AS.
     *
     * param uuid The UUID of the attestation to revoke.
     */
    function revoke(bytes32 uuid) external;

    /**
     * dev Attests to a specific AS using a provided EIP712 signature.
     *
     * param uuid The UUID of the attestation to revoke.
     * param attester The attesting account.
     * param v The recovery ID.
     * param r The x-coordinate of the nonce R.
     * param s The signature data.
     */
    function revokeByDelegation(
        bytes32 uuid,
        address attester,
        uint8 v,
        bytes32 r,
        bytes32 s
    ) external;

    /**
     * dev Returns an existing attestation by UUID.
     *
     * param uuid The UUID of the attestation to retrieve.
     *
     * return The attestation data members.
     */
    function getAttestation(bytes32 uuid)
        external
        view
        returns (Attestation memory);

    /**
     * dev Checks whether an attestation exists.
     *
     * param uuid The UUID of the attestation to retrieve.
     *
     * return Whether an attestation exists.
     */
    function isAttestationValid(bytes32 uuid) external view returns (bool);

    /**
     * dev Checks whether an attestation is active.
     *
     * param uuid The UUID of the attestation to retrieve.
     *
     * return Whether an attestation is active.
     */
    function isAttestationActive(bytes32 uuid) external view returns (bool);

    /**
     * dev Returns all received attestation UUIDs.
     *
     * param recipient The recipient of the attestation.
     * param schema The UUID of the AS.
     * param start The offset to start from.
     * param length The number of total members to retrieve.
     * param reverseOrder Whether the offset starts from the end and the data is returned in reverse.
     *
     * return An array of attestation UUIDs.
     */
    function getReceivedAttestationUUIDs(
        address recipient,
        bytes32 schema,
        uint256 start,
        uint256 length,
        bool reverseOrder
    ) external view returns (bytes32[] memory);

    /**
     * dev Returns the number of received attestation UUIDs.
     *
     * param recipient The recipient of the attestation.
     * param schema The UUID of the AS.
     *
     * return The number of attestations.
     */
    function getReceivedAttestationUUIDsCount(address recipient, bytes32 schema)
        external
        view
        returns (uint256);

    /**
     * dev Returns all sent attestation UUIDs.
     *
     * param attester The attesting account.
     * param schema The UUID of the AS.
     * param start The offset to start from.
     * param length The number of total members to retrieve.
     * param reverseOrder Whether the offset starts from the end and the data is returned in reverse.
     *
     * return An array of attestation UUIDs.
     */
    function getSentAttestationUUIDs(
        address attester,
        bytes32 schema,
        uint256 start,
        uint256 length,
        bool reverseOrder
    ) external view returns (bytes32[] memory);

    /**
     * dev Returns the number of sent attestation UUIDs.
     *
     * param recipient The recipient of the attestation.
     * param schema The UUID of the AS.
     *
     * return The number of attestations.
     */
    function getSentAttestationUUIDsCount(address recipient, bytes32 schema)
        external
        view
        returns (uint256);

    /**
     * dev Returns all attestations related to a specific attestation.
     *
     * param uuid The UUID of the attestation to retrieve.
     * param start The offset to start from.
     * param length The number of total members to retrieve.
     * param reverseOrder Whether the offset starts from the end and the data is returned in reverse.
     *
     * return An array of attestation UUIDs.
     */
    function getRelatedAttestationUUIDs(
        bytes32 uuid,
        uint256 start,
        uint256 length,
        bool reverseOrder
    ) external view returns (bytes32[] memory);

    /**
     * dev Returns the number of related attestation UUIDs.
     *
     * param uuid The UUID of the attestation to retrieve.
     *
     * return The number of related attestations.
     */
    function getRelatedAttestationUUIDsCount(bytes32 uuid)
        external
        view
        returns (uint256);

    /**
     * dev Returns all per-schema attestation UUIDs.
     *
     * param schema The UUID of the AS.
     * param start The offset to start from.
     * param length The number of total members to retrieve.
     * param reverseOrder Whether the offset starts from the end and the data is returned in reverse.
     *
     * return An array of attestation UUIDs.
     */
    function getSchemaAttestationUUIDs(
        bytes32 schema,
        uint256 start,
        uint256 length,
        bool reverseOrder
    ) external view returns (bytes32[] memory);

    /**
     * dev Returns the number of per-schema  attestation UUIDs.
     *
     * param schema The UUID of the AS.
     *
     * return The number of attestations.
     */
    function getSchemaAttestationUUIDsCount(bytes32 schema)
        external
        view
        returns (uint256);
}
